@@ -53,5 +53,5 @@ def embed_kernel(module=None, local_ns=None, **kwargs):
 
     app.kernel.user_module = module
     app.kernel.user_ns = local_ns
-    app.shell.set_completer_frame()
+    app.shell.set_completer_frame()  # type:ignore[union-attr]
     app.start()
